# 
#   MLCV Winter 2017
#   Homework Assignment 06 - Hidden Markov Models
#
#   Exercise 1c) Viterbi algorithm
#
###########################################################################

import numpy as np


def viterbi(Z, p_init, p_transition, p_sense, state_space, observation_space):
    """Recursively estimates the state sequence that maximizes p(X,Z|θ).

    Parameters
    ----------
    Z : array, (n_samples,)
        A sequence of observations encoded as integers in [0, n_observations).

    p_init : array, (n_states)
        The prior probability of each state.

    p_transition : array, (n_states, n_states)
        The transition probabilities.

    p_sense : array, (n_observations, n_states)
        The observation model.

    state_space : array, (n_states,)
        A "human-readable" definition of the state space.

    observation_space : array, (n_observations)
        A "human-readable" definition of the observation space.

    Returns
    -------
    path : array, (n_samples,)
        The most probable sequence of states that was followed.

    prob : float
        The probability of the returned path.

    """

    n_observations, n_states = p_sense.shape

    # Initialize
    print(f'\nObservation at t={0}: {observation_space[Z[0]]}')

    # δ(x) = p(z|x)*p(x)
    delta = p_sense[Z[0], :] * p_init

    # Normalize
    delta /= delta.sum()

    # Store the most probable paths to a dictionary
    paths = {state:[state] for state in range(n_states)}

    # Print some information
    header_fields = ['state', 'prev_state (max)', 'prob']
    fmt = '{:10} {:15} {:^10}'
    header = fmt.format(*header_fields)

    print(f'{header}\n{"-"*len(header)}')
    for state, prob in enumerate(delta):
        print(f'{str(state_space[state]):10} {"-":^15} {prob:>10.4f}')

    # Find (recursively) the most probable path
    for t in range(1, len(Z)):  # For each observation
        print(f'\nObservation at t={t}: {observation_space[Z[t]]}')

        delta_prev = delta.copy()  # Last posterior is the new prior
        paths_prev = paths.copy()

        print(f'{header}\n{"-"*len(header)}')

        # How can one end up at state s ?
        for state in range(n_states):  # For each state

            # Probabilities of getting to state in the last transition
            proba_last_transition = delta_prev * p_transition[:, state]

            # What was the most probable state at t-1 leading to s at t ?
            max_prev_state = np.argmax(proba_last_transition)  # ψ
            prob_max_prev_state = proba_last_transition[max_prev_state]

            print(f'{str(state_space[state]):10} '
                  f'{str(state_space[max_prev_state]):^15} '
                  f'{prob_max_prev_state:>10.4f}')

            delta[state] = p_sense[Z[t], state] * prob_max_prev_state
            paths[state] = paths_prev[max_prev_state] + [state]

        print(f'\nMost likely paths:')
        for state, path in paths.items():
            readable_path = [state_space[s] for s in path]
            print(f'{state_space[state]}: {readable_path}')

        # Rescale / Normalize
        delta /= delta.sum()

    # Return the most likely sequence and its probability
    max_prev_state = np.argmax(delta)
    prob_max_prev_state = delta[max_prev_state]
    paths = paths[max_prev_state]

    return paths, prob_max_prev_state


def main():
    # Initializations
    world = np.array([['G', 'R', 'B'],
                      ['R', 'B', 'G'],
                      ['B', 'G', 'R']])

    # Define the observation space
    observation_space = ['R', 'G', 'B']

    # Define the state space
    state_space = []
    for i in range(world.shape[0]):
        for j in range(world.shape[1]):
            state_space.append((i, j))

    state_id_map = {state: i for i, state in enumerate(state_space)}

    def is_valid_state(state):
        for dim in range(len(state)):
            if state[dim] < 0 or state[dim] >= world.shape[dim]:
                return False
        return True

    # Define the state transition matrix
    p_transition = np.zeros(shape=(len(state_space), len(state_space)))

    # Transitions: up, down, left, right
    moves = [(0, 1), (0, -1), (-1, 0), (1, 0)]
    moves = [np.array(move) for move in moves]

    # Random movement implies each move is equally probable
    p_move = 1. / len(moves)

    # Now consider the constraints of the state space
    for state_from, state in enumerate(state_space):
        # Try to move up, down, left, right
        for move in moves:
            new_state = tuple(state[d] + move[d] for d in range(len(state)))

            # If out of bounds, stay at the current state
            if not is_valid_state(new_state):
                new_state = state

            state_to = state_id_map[new_state]
            p_transition[state_from, state_to] += p_move

    print('Transition Matrix:')
    print(p_transition)

    # Define observation model p(z|x) from exercise 1
    # Each column is the probability to sense colors z, given color x
    observation_model = np.array([[0.8, 0.1, 0.1],
                                  [0.1, 0.6, 0.2],
                                  [0.1, 0.3, 0.7]])

    # Define observation model for the new state space
    # Each column is the probability to sense colors z, given state x
    p_sense = np.zeros(shape=(len(observation_space), len(state_space)))
    for state_id, state in enumerate(state_space):
        state_color = world[state[0], state[1]]
        observation_id = observation_space.index(state_color)
        p_sense[:, state_id] = observation_model[:, observation_id]

    print('\nEmission Matrix:')
    print(p_sense)

    # Define initial state probabilities p(x_{t=0})
    p_init = np.ones(len(state_space)) / len(state_space)

    # The observed sequence
    Z = ['R', 'G', 'G']
    observed_sequence = np.array([observation_space.index(z) for z in Z])

    path_vit, prob_vit = viterbi(observed_sequence, p_init, p_transition,
                                 p_sense, state_space, observation_space)

    path_vit_readable = [state_space[state_id] for state_id in path_vit]
    print(f'\nThe most likely taken path with probability {prob_vit:.4f} is '
          f'{path_vit_readable}.')


if __name__ == '__main__':
    main()